#include <unistd.h>
#include "dbus_serialize.hpp"
#include "dbus_print.hpp"
#include "utils.hpp"

void send_dbus_message(const int fd, const DBusMessage& message) {
  std::vector<uint32_t> arraySizes;
  SerializerInitArraySizes s0(arraySizes);
  message.serialize(s0);

  size_t size = s0.getPos();
  std::unique_ptr<char[]> buf(new char[size]);
  SerializeToBuffer<LittleEndian> s1(arraySizes, buf.get());
  message.serialize(s1);

  write(fd, buf.get(), size);
}

std::unique_ptr<DBusMessage> receive_dbus_message(const int fd) {
  std::unique_ptr<DBusMessage> message;
  Parse p(DBusMessage::parseLE(message));

  while (true) {
    char buf[256];
    size_t required = p.maxRequiredBytes();
    if (required == 0) {
      return message;
    }
    if (required > sizeof(buf)) {
      required = sizeof(buf);
    }
    const ssize_t n = read(fd, buf, required);
    if (n < 0 || size_t(n) < required) {
      throw ParseError(p.getPos(), _s("No more input. n=") + std::to_string(n));
    }
#if 0
    printf("%ld: ", required);
    for (size_t i = 0; i < required; i++) {
      printf("%.2x", (unsigned int)(unsigned char)buf[i]);
    }
    printf("\n");
#endif
    p.parse(buf, required);
  }
}

void print_dbus_object(const int fd, const DBusObject& obj) {
  PrinterFD printFD(fd, 16, 2);
  obj.print(printFD);
  printFD.printNewline(0);
}

void print_dbus_message(const int fd, const DBusMessage& message) {
  PrinterFD printFD(fd, 16, 2);
  message.print(printFD, 0);
  printFD.printNewline(0);
}

void dbus_send_hello(const int fd) {
  std::unique_ptr<DBusMessageBody> body = DBusMessageBody::mk0();
  const size_t bodySize = body->serializedSize();

  std::unique_ptr<DBusObject> header =
    DBusObjectStruct::mk(
      _vec<std::unique_ptr<DBusObject>>(
        // Header
        DBusObjectChar::mk('l'), // Little endian
        DBusObjectChar::mk(MSGTYPE_METHOD_CALL),
        DBusObjectChar::mk(MSGFLAGS_EMPTY),
        DBusObjectChar::mk(1), // Major protocol version
        DBusObjectUint32::mk(bodySize), // body_len_unsigned
        DBusObjectUint32::mk(0x1001), // serial number
        DBusObjectArray::mk1(
          _vec<std::unique_ptr<DBusObject>>(
            DBusHeaderField::mk(
              MSGHDR_PATH,
              DBusObjectVariant::mk(
                DBusObjectPath::mk(_s("/org/freedesktop/DBus"))
              )
            ),
            DBusHeaderField::mk(
              MSGHDR_INTERFACE,
              DBusObjectVariant::mk(
                DBusObjectString::mk(_s("org.freedesktop.DBus"))
              )
            ),
            DBusHeaderField::mk(
              MSGHDR_DESTINATION,
              DBusObjectVariant::mk(
                DBusObjectString::mk(_s("org.freedesktop.DBus"))
              )
            ),
            DBusHeaderField::mk(
              MSGHDR_MEMBER,
              DBusObjectVariant::mk(
                DBusObjectString::mk(_s("Hello"))
              )
            ),
            DBusHeaderField::mk(
              MSGHDR_REPLY_SERIAL,
              DBusObjectVariant::mk(
                DBusObjectUint32::mk(0x2001)
              )
            )
          )
        )
      )
    );

  const DBusMessage message(std::move(header), std::move(body));

  send_dbus_message(fd, message);
}
